@@ -22,20 +22,23 @@
 Kaldi engine classes
 """
 
-import collections, logging, os, subprocess, sys, threading, time
+import collections, functools, logging, os, sys, time
 
 from packaging.version import Version
-from six import integer_types, string_types, print_, reraise
+from six import PY2, integer_types, string_types, print_, reraise
 from six.moves import zip
 import kaldi_active_grammar
-from kaldi_active_grammar       import KaldiAgfNNet3Decoder, KaldiError
+from kaldi_active_grammar       import KaldiAgfNNet3Decoder, KaldiError, KaldiRule
 
-from ..base                     import (EngineBase, EngineError, MimicFailure,
+from ..base                     import (EngineBase,
+                                        EngineError, CompilerError,
+                                        MimicFailure,
                                         DelegateTimerManager,
                                         DelegateTimerManagerInterface,
                                         DictationContainerBase,
                                         GrammarWrapperBase)
 from .audio                     import MicAudio, VADAudio, AudioStore, WavAudio
+from .dictation                 import user_dictation_list, user_dictation_dictlist
 from .recobs                    import KaldiRecObsManager
 from .testing                   import debug_timer
 from dragonfly.grammar.state    import State
@@ -63,10 +66,10 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
 
     #-----------------------------------------------------------------------
 
-    def __init__(self, model_dir=None, tmp_dir=None,
-        input_device_index=None, audio_self_threaded=True,
+    def __init__(self, model_dir=None, tmp_dir=None, input_device_index=None,
+        audio_input_device=None, audio_self_threaded=True, audio_auto_reconnect=True, audio_reconnect_callback=None,
         retain_dir=None, retain_audio=None, retain_metadata=None, retain_approval_func=None,
-        vad_aggressiveness=3, vad_padding_start_ms=150, vad_padding_end_ms=150, vad_complex_padding_end_ms=500,
+        vad_aggressiveness=3, vad_padding_start_ms=150, vad_padding_end_ms=200, vad_complex_padding_end_ms=600,
         auto_add_to_user_lexicon=True, lazy_compilation=True, invalidate_cache=False,
         expected_error_rate_threshold=None,
         alternative_dictation=None, cloud_dictation_lang='en-US',
@@ -100,18 +103,30 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
                 self._log.warning("%s: Enabling logging of actions execution to avoid bug processing keyboard actions on Windows", self)
                 action_exec_logger.setLevel(logging.DEBUG)
 
-        if not (isinstance(retain_dir, string_types) or (retain_dir is None)):
-            self._log.error("Invalid retain_dir: %r" % retain_dir)
-            retain_dir = None
+        # Handle engine parameters
+        if input_device_index is not None:
+            if audio_input_device is not None:
+                raise ValueError("Cannot set both input_device_index and audio_input_device")
+            self._log.warning("%s: input_device_index is deprecated; please use audio_input_device", self)
+            audio_input_device = int(input_device_index)
+        if audio_input_device not in (None, False) and not isinstance(audio_input_device, (int, string_types)):
+            raise TypeError("Invalid audio_input_device not int or string: %r" % (audio_input_device,))
+        if audio_reconnect_callback is not None and not callable(audio_reconnect_callback):
+            raise TypeError("Invalid audio_reconnect_callback not callable: %r" % (audio_reconnect_callback,))
+        if retain_dir is not None and not isinstance(retain_dir, string_types):
+            raise TypeError("Invalid retain_dir not string: %r" % (retain_dir,))
         if retain_audio and not retain_dir:
-            self._log.error("retain_audio=True requires retain_dir to be set; making retain_audio=False instead")
-            retain_audio = False
+            raise ValueError("retain_audio=True requires retain_dir to be set")
+        if retain_approval_func is not None and not callable(retain_approval_func):
+            raise TypeError("Invalid retain_approval_func not callable: %r" % (retain_approval_func,))
 
         self._options = dict(
             model_dir = model_dir,
             tmp_dir = tmp_dir,
-            input_device_index = input_device_index,
+            audio_input_device = audio_input_device,
             audio_self_threaded = bool(audio_self_threaded),
+            audio_auto_reconnect = bool(audio_auto_reconnect),
+            audio_reconnect_callback = audio_reconnect_callback,
             retain_dir = retain_dir,
             retain_audio = bool(retain_audio) if retain_audio is not None else bool(retain_dir),
             retain_metadata = bool(retain_metadata) if retain_metadata is not None else bool(retain_dir),
@@ -129,14 +144,22 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
             decoder_init_config = dict(decoder_init_config) if decoder_init_config else None,
         )
 
-        self._compiler = None
-        self._decoder = None
-        self._audio = None
-        self.audio_store = None
+        # Setup
+        self._reset_state()
         self._recognition_observer_manager = KaldiRecObsManager(self)
         self._timer_manager = DelegateTimerManager(0.02, self)
 
+    def _reset_state(self):
+        self._compiler = None
+        self._decoder = None
+        self._audio = None
+        self._audio_iter = None
+        self.audio_store = None
+
+        self._any_exclusive_grammars = False
         self._saving_adaptation_state = False
+        self._ignore_current_phrase = False
+        self._in_phrase = False
         self._doing_recognition = False
         self._deferred_disconnect = False
 
@@ -144,10 +167,10 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
         """ Connect to back-end SR engine. """
         if self._decoder:
             return
+        self._reset_state()
 
         self._log.info("Loading Kaldi-Active-Grammar v%s in process %s." % (kaldi_active_grammar.__version__, os.getpid()))
         self._log.info("Kaldi options: %s" % self._options)
-        # subprocess.call(['vsjitdebugger', '-p', str(os.getpid())]); time.sleep(5)
 
         self._compiler = KaldiCompiler(self._options['model_dir'], tmp_dir=self._options['tmp_dir'],
             auto_add_to_user_lexicon=self._options['auto_add_to_user_lexicon'],
@@ -165,38 +188,34 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
             config=self._options['decoder_init_config'],)
         self._compiler.decoder = self._decoder
 
-        self._audio = VADAudio(
-            aggressiveness=self._options['vad_aggressiveness'],
-            start=False,
-            input_device_index=self._options['input_device_index'],
-            self_threaded=self._options['audio_self_threaded'],
-            )
-        self._audio_iter = self._audio.vad_collector(nowait=True,
-            start_window_ms=self._options['vad_padding_start_ms'],
-            end_window_ms=self._options['vad_padding_end_ms'],
-            complex_end_window_ms=self._options['vad_complex_padding_end_ms'],
-            )
-        self.audio_store = AudioStore(self._audio, maxlen=(1 if self._options['retain_dir'] else 0),
-            save_dir=self._options['retain_dir'], save_audio=self._options['retain_audio'], save_metadata=self._options['retain_metadata'],
-            retain_approval_func=self._options['retain_approval_func'])
-
-        self._any_exclusive_grammars = False
-        self._in_phrase = False
-        self._ignore_current_phrase = False
+        if self._options['audio_input_device'] is not False:
+            self._audio = VADAudio(
+                aggressiveness=self._options['vad_aggressiveness'],
+                start=False,
+                input_device=self._options['audio_input_device'],
+                self_threaded=self._options['audio_self_threaded'],
+                reconnect_callback=self._options['audio_reconnect_callback'],
+                )
+            self._audio_iter = self._audio.vad_collector(nowait=True,
+                audio_auto_reconnect=self._options['audio_auto_reconnect'],
+                start_window_ms=self._options['vad_padding_start_ms'],
+                end_window_ms=self._options['vad_padding_end_ms'],
+                complex_end_window_ms=self._options['vad_complex_padding_end_ms'],
+                )
+            self.audio_store = AudioStore(self._audio, maxlen=(1 if self._options['retain_dir'] else 0),
+                save_dir=self._options['retain_dir'], save_audio=self._options['retain_audio'], save_metadata=self._options['retain_metadata'],
+                retain_approval_func=self._options['retain_approval_func'])
 
     def disconnect(self):
         """ Disconnect from back-end SR engine. Exits from ``do_recognition()``. """
         if self._doing_recognition:
             self._deferred_disconnect = True
         else:
-            self._deferred_disconnect = False
             if self._audio:
                 self._audio.destroy()
-                self._audio = None
-                self._audio_iter = None
-                self.audio_store = None
-            self._compiler = None
-            self._decoder = None
+            if self.audio_store:
+                self.audio_store.save_all()
+            self._reset_state()
             self._grammar_wrappers = {}  # From EngineBase
 
     def print_mic_list(self):
@@ -214,7 +233,8 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
         kaldi_rule_by_rule_dict = self._compiler.compile_grammar(grammar, self)
         wrapper = GrammarWrapper(grammar, kaldi_rule_by_rule_dict, self,
                                  self._recognition_observer_manager)
-        for kaldi_rule in kaldi_rule_by_rule_dict.values():
+        for (rule, kaldi_rule) in kaldi_rule_by_rule_dict.items():
+            kaldi_rule.active = bool(rule.active)  # Initialize to correct activity
             kaldi_rule.load(lazy=self._compiler.lazy_compilation)
 
         return wrapper
@@ -291,7 +311,13 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
 
     def prepare_for_recognition(self):
         """ Can be called optionally before ``do_recognition()`` to speed up its starting of active recognition. """
-        self._compiler.prepare_for_recognition()
+        try:
+            self._compiler.prepare_for_recognition()
+        except KaldiError as e:
+            if len(e.args) >= 2 and isinstance(e.args[1], KaldiRule):
+                kaldi_rule = e.args[1]
+                raise self._compiler.make_compiler_error_for_kaldi_rule(kaldi_rule)
+            raise
 
     def _do_recognition(self, timeout=None, single=False, audio_iter=None):
         """
@@ -301,16 +327,18 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
         self._log.debug("do_recognition: timeout %s" % timeout)
         if not self._decoder:
             raise EngineError("Cannot recognize before connect()")
+        if audio_iter is None and self._audio is None:
+            raise EngineError("No audio input")
         self._doing_recognition = True
+        self._in_phrase = False
+        self._ignore_current_phrase = False
+        in_complex = False
+        end_time = None
+        timed_out = False
 
         try:
             self.prepare_for_recognition()
 
-            self._in_phrase = False
-            self._ignore_current_phrase = False
-            in_complex = False
-            timed_out = False
-
             if timeout != None:
                 end_time = time.time() + timeout
                 timed_out = True
@@ -322,7 +350,7 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
             next(audio_iter)  # Prime the audio iterator
 
             # Loop until timeout (if set) or until disconnect() is called.
-            while (not self._deferred_disconnect) and ((not timeout) or (time.time() < end_time)):
+            while (not self._deferred_disconnect) and ((not end_time) or (time.time() < end_time)):
                 self.prepare_for_recognition()
                 block = audio_iter.send(in_complex)
 
@@ -395,19 +423,33 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
 
         finally:
             self._doing_recognition = False
-            if audio_iter == self._audio_iter and self._audio:
-                self._audio.stop()
-            if self.audio_store:
-                self.audio_store.save_all()
+            if (audio_iter == self._audio_iter) and self._audio:
+                try:
+                    self._audio.stop()  # We started the audio above, so we should stop it
+                except Exception:
+                    self._log.exception("Error stopping audio")
+            if self._deferred_disconnect:
+                self.disconnect()
 
         return not timed_out
 
     in_phrase = property(lambda self: self._in_phrase,
         doc="Whether or not the engine is currently in the middle of hearing a phrase from the user.")
 
-    def recognize_wave_file(self, filename, **kwargs):
-        """ Does recognition on given wave file, treating it as a single utterance (without VAD), then returns. """
-        self.do_recognition(audio_iter=WavAudio.read_file(filename), **kwargs)
+    def recognize_wave_file(self, filename, realtime=False, **kwargs):
+        """
+            Does recognition on given wave file, treating it as a single
+            utterance (without VAD), then returns.
+        """
+        self.do_recognition(audio_iter=WavAudio.read_file(filename, realtime=realtime), **kwargs)
+
+    def recognize_wave_file_as_stream(self, filename, realtime=False, **kwargs):
+        """
+            Does recognition on given wave file, treating it as a stream and
+            processing it with VAD to break it into multiple utterances (as with
+            normal microphone audio input), then returns.
+        """
+        self.do_recognition(audio_iter=WavAudio.read_file_with_vad(filename, realtime=realtime), **kwargs)
 
     def ignore_current_phrase(self):
         """
@@ -436,6 +478,31 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
     def reset_adaptation_state(self):
         self._decoder.reset_adaptation_state()
 
+    def add_word_list_to_user_dictation(self, word_list):
+        """ Make UserDictation elements able to recognize each item of given
+            list of strings *word_list*. Note: all characters will be converted
+            to lowercase, and recognized as such. """
+        word_list = [str(word).lower() for word in word_list]
+        word_list = [word for word in word_list
+            if word not in user_dictation_list]
+        # if any((word != word.lower()) for word in word_list):
+        #     raise ValueError("Cannot recognize words with uppercase")
+        user_dictation_list.extend(word_list)
+
+    def add_word_dict_to_user_dictation(self, word_dict):
+        """ Make UserDictation elements able to recognize each item of given
+            dict of strings *word_dict*. The key is the "spoken form" (which is
+            recognized), and the value is the "written form" (which is returned
+            as the text in the UserDictation element). Note: all characters in
+            the keys will be converted to lowercase, but the values are returned
+            as text verbatim. """
+        word_dict = {str(key).lower(): str(value) for key, value in word_dict.items()}
+        word_dict = {key: value for key, value in word_dict.items()
+            if key not in user_dictation_dictlist.keys()}
+        # if any((word != word.lower()) for word in word_dict.keys()):
+        #     raise ValueError("Cannot recognize words with uppercase")
+        user_dictation_dictlist.update(word_dict)
+
     #-----------------------------------------------------------------------
     # Internal processing methods.
 
@@ -464,7 +531,7 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
         self._active_kaldi_rules = set()
         self._kaldi_rules_activity = [False] * self._compiler.num_kaldi_rules
         for grammar_wrapper in self._iter_all_grammar_wrappers_dynamically():
-            if grammar_wrapper.active and (not self._any_exclusive_grammars or (self._any_exclusive_grammars and grammar_wrapper.exclusive)):
+            if grammar_wrapper.active and (not self._any_exclusive_grammars or grammar_wrapper.exclusive):
                 for kaldi_rule in grammar_wrapper.kaldi_rule_by_rule_dict.values():
                     if kaldi_rule.active:
                         self._active_kaldi_rules.add(kaldi_rule)
@@ -514,7 +581,7 @@ class KaldiEngine(EngineBase, DelegateTimerManagerInterface):
 
             if self._log.isEnabledFor(12):
                 try:
-                    self._log.log(12, "Alignment (word,time,length): %s" % self._decoder.get_word_align(output))
+                    self._log.log(12, "Alignment (word,time,length): %s", self._decoder.get_word_align(output))
                 except KaldiError as e:
                     self._log.warning("Exception logging word alignment")
 
@@ -558,10 +625,11 @@ class Recognition(object):
     def construct_empty(cls, engine):
         return cls(engine, kaldi_rule=None, words=())
 
-    def __del__(self):
-        # Exactly one of process() or fail() should be called by someone!
-        if not self.finalized:
-            self.engine._log.warn("%s not finalized!", self)
+    # def __del__(self):
+    #     # Exactly one of process() or fail() should be called by someone!
+    #     if not self.finalized:
+    #         self.engine._log.warning("%s not finalized!", self)
+    #     # Note: this can be generated spurriously upon exit or testing
 
     def process(self, expected_error_rate=None, confidence=None):
         if expected_error_rate is not None: self.expected_error_rate = expected_error_rate
@@ -601,6 +669,8 @@ class GrammarWrapper(GrammarWrapperBase):
         rule = recognition.kaldi_rule.parent_rule
         words_are_dictation_mask = recognition.words_are_dictation_mask
         try:
+            assert (rule.active and rule.exported), "Kaldi engine should only ever return the correct rule"
+
             # Prepare the words and rule names for the element parsers
             rule_names = (rule.name,) + (('dgndictation',) if any(words_are_dictation_mask) else ())
             words_rules = tuple((word, 0 if not is_dictation else 1)
